#include "stdafx.h"
#include "clf_resources.h"
#include "clf_image.h"
#include "clf_mesh.h"
#include "glew.h"

// comment the following line to draw the geometry without vertex buffers
#define USE_VBOS

clf_mesh :: clf_mesh()
{
	// Set Pointers To NULL
	m_pVertices = NULL;
	m_pTexCoords = NULL;
	m_nVertexCount = 0;
	m_nVBOVertices = m_nVBOTexCoords = m_nTextureId = 0;
}

clf_mesh :: ~clf_mesh()
{
	// Delete VBOs
	unsigned int nBuffers[2] = { m_nVBOVertices, m_nVBOTexCoords };
	glDeleteBuffers( 2, nBuffers );						// Free The Memory

	// Delete Data
	if( m_pVertices )											// Deallocate Vertex Data
		delete [] m_pVertices;
	m_pVertices = NULL;
	if( m_pTexCoords )											// Deallocate Texture Coord Data
		delete [] m_pTexCoords;
	m_pTexCoords = NULL;

	m_nTextureId = 0;
}

bool clf_mesh :: load_height_map( char* file_name, float height_scale, float length_scale, float resolution )
{
	// load image data
	clf_targa_image imageData;
	// Error-Checking
	if (!imageData.load(get_resource_file_name (file_name, RES_TEX)))
		return false;

	int vert, nIndex;									// Create Variables
	float nX, nZ;
	float flX, flZ;

	int l_w = imageData.get_width();
	int l_h = imageData.get_height();

	if (resolution < 1.f)
		resolution = 1.f;

	int v_c = (int)(l_w / resolution);
	int vcs = (int)(v_c * resolution);
	while (vcs < l_w)
	{
		v_c ++;
		vcs = (int)(v_c * resolution);
	}

	// Generate Vertex Field
	m_nVertexCount = v_c*v_c*6;
	m_pVertices = new clf_vector3D[m_nVertexCount*2];			// allocate vertex and normals data
	m_pNormals = &m_pVertices[m_nVertexCount];
	m_pTexCoords = new clf_point2D[m_nVertexCount];				// allocate texture coordinates data
	nIndex = 0;
	for( nZ = 0; nZ < l_h; nZ +=  resolution )
	{
		for( nX = 0; nX < l_w; nX += resolution )
		{
			// generate two triangles for each quad
			for( vert = 0; vert < 6; vert++ )
			{
				// calculate X,Z image coordinates
				flX = nX;
				if ( vert == 1 || vert == 2 || vert == 5 )
					flX += resolution;
				flZ = nZ;
				if ( vert == 2 || vert == 4 || vert == 5 )
					flZ += resolution;

				// calculate X,Z world coordinates
				m_pVertices[nIndex].x = (flX - ( l_w / 2 ))*length_scale;
				m_pVertices[nIndex].z = (flZ - ( l_h / 2 ))*length_scale;
				// and finally calculate Y coodrinate based on alpha value
				m_pVertices[nIndex].y = get_alpha( imageData.get_image(), l_w, l_h, dti(flX), dti(flZ) ) *  height_scale;

				// set texture coordinates
				m_pTexCoords[nIndex].x = flX / l_w;
				m_pTexCoords[nIndex].y = flZ / l_h;

				// when the triangle is complete calculate normal vector
				if (vert == 2) // first triangle
				{
					m_pNormals[nIndex-2] = m_pNormals[nIndex-1] = m_pNormals[nIndex] = calc_face_norm(nIndex-1,nIndex,nIndex-2);
				}
				else if (vert == 5) // second triangle
				{
					m_pNormals[nIndex-2] = m_pNormals[nIndex-1] = m_pNormals[nIndex] = calc_face_norm(nIndex,nIndex-1,nIndex-2);
				}

				// Increment Our Index
				nIndex++;
			}
		}
	}

	// load The Texture Into OpenGL, we assume the height map is also the texture
	m_nTextureId = clf_loadTexture(file_name);

	build_buffers();									// Build The VBOs

	return true;
}

void clf_mesh :: build_buffers()
{
	// Generate And Bind The Vertex Buffer
	glGenBuffers( 1, &m_nVBOVertices );							// Get A Valid Name
	glBindBuffer( GL_ARRAY_BUFFER, m_nVBOVertices );			// Bind The Buffer 
	// load The Data
	DWORD len = m_nVertexCount*sizeof(clf_vector3D);
	// allocate a buffer for the vertices and the normals, filled with seros
	glBufferData( GL_ARRAY_BUFFER, len*2, 0, GL_STATIC_DRAW ); // (vertices and normals)
	// copy the vertices in the first half
	glBufferSubData(GL_ARRAY_BUFFER, 0, len, m_pVertices);
	// and the normals in the rest of the buffer
	glBufferSubData(GL_ARRAY_BUFFER, len, len, m_pNormals);
	glBindBuffer( GL_ARRAY_BUFFER, 0 );			// Release The Buffer 

	if (m_nTextureId > 0)
	{
		// Generate And Bind The Texture Coordinate Buffer
		glGenBuffers( 1, &m_nVBOTexCoords );							// Get A Valid Name
		glBindBuffer( GL_ARRAY_BUFFER, m_nVBOTexCoords );		// Bind The Buffer
		// load The Data
		glBufferData( GL_ARRAY_BUFFER, m_nVertexCount*sizeof(clf_point2D), m_pTexCoords, GL_STATIC_DRAW );

		glBindBuffer( GL_ARRAY_BUFFER, 0 );		// Release The Buffer
	}
}

#ifndef USE_VBOS
void clf_mesh::render()
{
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, m_nTextureId);
	int i;
	glBegin(GL_TRIANGLES);
	for (i=0; i<m_nVertexCount; i+=3)
	{
		//int j = i/3;
		glNormal3f(m_pNormals[i].x, m_pNormals[i].y, m_pNormals[i].z);
		glVertex3f(m_pVertices[i].x, m_pVertices[i].y, m_pVertices[i].z);
		glTexCoord2f(m_pTexCoords[i].x, m_pTexCoords[i].y);
		//glNormal3f(m_pNormals[i].x, m_pNormals[i].y, m_pNormals[i].z);
		glVertex3f(m_pVertices[i+1].x, m_pVertices[i+1].y, m_pVertices[i+1].z);
		glTexCoord2f(m_pTexCoords[i+1].x, m_pTexCoords[i+2].y);
		//glNormal3f(m_pNormals[i].x, m_pNormals[i].y, m_pNormals[i].z);
		glVertex3f(m_pVertices[i+2].x, m_pVertices[i+2].y, m_pVertices[i+2].z);
		glTexCoord2f(m_pTexCoords[i+2].x, m_pTexCoords[i+2].y);
	}
	glEnd();
}
#else
void clf_mesh::render()
{
	glEnable(GL_TEXTURE_2D);
	glEnableClientState( GL_VERTEX_ARRAY );						// Enable Vertex Arrays
	glEnableClientState(GL_NORMAL_ARRAY);
	if (m_nTextureId > 0)
		glEnableClientState( GL_TEXTURE_COORD_ARRAY );				// Enable Texture Coord Arrays

	// Set Pointers To Our Data
	glBindBuffer( GL_ARRAY_BUFFER, m_nVBOVertices );
	glNormalPointer(GL_FLOAT, sizeof(clf_vector3D), (void*)(m_nVertexCount*sizeof(clf_vector3D)));
	glVertexPointer( 3, GL_FLOAT, sizeof(clf_vector3D), (char *) NULL );		// Set The Vertex Pointer To The Vertex Buffer

	if (m_nTextureId > 0)
	{
		glBindBuffer( GL_ARRAY_BUFFER, m_nVBOTexCoords );
		glTexCoordPointer( 2, GL_FLOAT, sizeof(clf_point2D), (char *) NULL );		// Set The TexCoord Pointer To The TexCoord Buffer
		glBindTexture( GL_TEXTURE_2D, m_nTextureId );				// Bind The Texture
	}

	// render
	glDrawArrays( GL_TRIANGLES, 0, m_nVertexCount );	// Draw All Of The Triangles At Once

	// Disable Pointers
	glDisableClientState( GL_VERTEX_ARRAY );					// Disable Vertex Arrays
	glDisableClientState(GL_NORMAL_ARRAY);	// disable normal arrays
	if (m_nTextureId > 0)
		glDisableClientState( GL_TEXTURE_COORD_ARRAY );				// Disable Texture Coord Arrays
}
#endif

clf_vector3D clf_mesh :: calc_face_norm(int idx1, int idx2, int idx3)
{
	clf_vector3D& iv1 = m_pVertices[idx1];
	clf_vector3D& iv2 = m_pVertices[idx2];
	clf_vector3D& iv3 = m_pVertices[idx3];

	clf_vector3D v1(iv2.x - iv1.x, iv2.y - iv1.y, iv2.z - iv1.z);
	clf_vector3D v2(iv3.x - iv1.x, iv3.y - iv1.y, iv3.z - iv1.z);

	clf_vector3D norm = crossProduct(v2, v1);
	norm.normalize();

	return clf_vector3D(norm.x, norm.y, norm.z);
}
